% -------------------------------------------------------------------------
%      Topology optimization of compliant mechanisms with distributed 
%   compliance (hinge-free compliant mechanisms) by using stiffness and 
%        adaptive volume constraints instead of stress constraints
%
%                       Stephanie Seltmann, Alexander Hasse
%                           Mechanism and Machine Theory
% -------------------------------------------------------------------------

% *************************** Disclaimer *********************************
% The authors reserve all rights but do not guarantee that the code is
% free from errors. In addition, they are in no case liable for any event 
% caused by the use of the program.  
% ************************************************************************
clear all
%%%%%%%%%%%%%%%%%%%%%%Input of the required parameters%%%%%%%%%%%%%%%%%%%%%
% the numbers in brackets correspond to equations from the paper
%Parametrization
nelx = 120;	%Number of elements in horizontal direction
nely = 60;	%Number of elements in vertical direction
%Material
E = 210000;   %Young's modulus
nu1 = 0.3;   %Poisson's ratio
%boundary conditions and desired deformation mode
fixeddofs = horzcat(1:42,122:122:14762);	%fixed degrees of freedom
phi = [1 -1];	%desired deformation mode as row vector 
modedofs = [121 14761];	%master-DoFs
mu = 200;   %Limit for primary stiffness
%Optimization parameters according to table 1
rmin = 1.2;
Vt = 1;
eta = 4;
delta = 0.0002;
gamma = 0.999999;
epsilon = 5;
n=500;
Gmin = 0.0001;
xmin = Gmin*ones(nelx*nely,1);
xmax = ones(nelx*nely,1);
nu=0.002;
AnzM = 1;   %Number of modes used to stabilize optimization procedure
NumIte = 50000;	%Maximum number of iterations

%%%%%%%%%%%%%%%%%% Calculation of constant matrices %%%%%%%%%%%%%%%%%%%%%%%
k2 = [1/2-nu1/6 1/8+nu1/8 -1/4-nu1/12 -1/8+3*nu1/8 -1/4+nu1/12 ...
    -1/8-nu1/8 nu1/6 1/8-3*nu1/8];
KE = E/(1-nu1^2)*[ k2(1) k2(2) k2(3) k2(4) k2(5) k2(6) k2(7) k2(8)
	k2(2) k2(1) k2(8) k2(7) k2(6) k2(5) k2(4) k2(3)
	k2(3) k2(8) k2(1) k2(6) k2(7) k2(4) k2(5) k2(2)
	k2(4) k2(7) k2(6) k2(1) k2(8) k2(3) k2(2) k2(5)
	k2(5) k2(6) k2(7) k2(8) k2(1) k2(2) k2(3) k2(4)
	k2(6) k2(5) k2(4) k2(3) k2(2) k2(1) k2(8) k2(7)
	k2(7) k2(4) k2(5) k2(2) k2(3) k2(8) k2(1) k2(6)
	k2(8) k2(3) k2(2) k2(5) k2(4) k2(7) k2(6) k2(1)];           
nodenrs = reshape(1:(1+nelx)*(1+nely),1+nely,1+nelx);
edofVec = reshape(2*nodenrs(1:end-1,1:end-1)+1,nelx*nely,1);
edofMat = repmat(edofVec,1,8)+repmat([0 1 2*nely+[2 3 0 1] -2 -1],...
    nelx*nely,1);
iK = reshape(kron(edofMat,ones(8,1))',64*nelx*nely,1);
jK = reshape(kron(edofMat,ones(1,8))',64*nelx*nely,1);
elements = length(edofVec);
standard = repmat(KE,1,elements);
d3 = repmat(KE,1,1,elements);
%Subdivision of the degrees of freedom
alldofs = 1:length(nodenrs(:,1))*length(nodenrs(1,:))*2;
freedofs = setdiff(alldofs,fixeddofs);
modefree  = setdiff(freedofs,modedofs);
%Normalizing the desired deformation mode
phi = ((1/sqrt(phi*phi')).*phi)';
%Prepare filter matrices
iH = ones(nelx*nely*(2*(ceil(rmin)-1)+1)^2,1);
jH = ones(size(iH));
sH = zeros(size(iH));
k1 = 0;
for i1 = 1:nelx
	for j1 = 1:nely
        e1 = (i1-1)*nely+j1;
        for i2 = max(i1-(ceil(rmin)-1),1):min(i1+(ceil(rmin)-1),nelx)
            for j2 = max(j1-(ceil(rmin)-1),1):min(j1+(ceil(rmin)-1),nely)
                e2 = (i2-1)*nely+j2;
                k1 = k1+1;
                iH(k1) = e1;
                jH(k1) = e2;
                sH(k1) = max(0,rmin-sqrt((i1-i2)^2+(j1-j2)^2));
            end
        end
	end
end
H = sparse(iH,jH,sH);
Hs = sum(H,2);
%Start vector
x0 = random('unif',xmin,xmax);
%Set options optimization algorithm
options = optimoptions(@linprog,'Algorithm','dual-simplex','MaxIter',...
    3000,'Preprocess','none','Tolcon',1e-3,'Display','off');

%%%%%%%%%%%%%%%%%%%%% Global iteration procedure %%%%%%%%%%%%%%%%%%%%%%%%%%
for y = 1:NumIte
	%%%%%%%%%%%% Calculate objective function and restrictions %%%%%%%%%%%%
	[A,f,b] = Displacement_distribution(x0,alldofs,phi,modedofs,AnzM,...
        edofMat,modefree,iK,jK,standard,d3,elements,Vt,mu); 
	%Filter objective function values (20)
	fnew(:) = H*(x0(:).*f(:))./Hs./x0(:);
	%%%%%%%%%% Sub-problem 2: updating the design variables %%%%%%%%%%%%%%%
	[x,fval,exitflag,output,lambda] = linprog(fnew,A,b,[],[],xmin,xmax,...
        [],options);
	if exitflag==-2
        disp(['no solution can be found with these input parameters and'...
        'this starting vector']);
        break;
	end
    x0 = x;
	%Stabilization (15)
	xmax = x0+nu;
	xmax(xmax>1) = 1;
	xmin = x0-nu;
	xmin(xmin<Gmin) = Gmin;
	%Check whether the upper limit of the primary stiffness constraint
    %has been reached within some tolerance (23)-(24)
	if A(1,:)*x>gamma*mu
        Vt = Vt*(1-delta);
	else
        Vt = Vt*(1+delta); 
	end
	%Convergence criterion (25)-(26)
    Vs(y,1)=sum(x0);
	if y>n
        Vs_con = Vs(y-(n-1):y,1);
        if Vs_con(2:end)<=Vs_con(1)*(1+delta)^epsilon & ...
                Vs_con(2:end)>=Vs_con(1)*(1-delta)^epsilon
            break;
        end
    end
    colormap(gray); imagesc(1-reshape(x,nely,nelx)); caxis([0 1]); ...
        axis equal; axis off; drawnow;
end